/*++

Module Name:

    driver.c

Abstract:

    This file contains the driver entry points and callbacks.

Environment:

    Kernel-mode Driver Framework

--*/

#include "driver.h"
#include "driver.tmh"
#include "wdmguid.h"
TBarData BarFromResource(CM_PARTIAL_RESOURCE_DESCRIPTOR& Resource);
void DelBar(TBarData& Bar); 

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#pragma alloc_text (PAGE, AIOWDM4EvtDeviceAdd)
#pragma alloc_text (PAGE, AIOWDM4EvtDriverContextCleanup)
#pragma alloc_text (PAGE, AIOWDM4MapHWResources)
#pragma alloc_text (PAGE, AIOWDM4UnmapHWResources) 
#pragma alloc_text (PAGE, AIOWDM4EvtPrepareHardware) 
#pragma alloc_text (PAGE, AIOWDM4GetDeviceInformation)
#pragma alloc_text (PAGE, AIOWDM4CreateIRQQueue)

#endif
PDEVICE_OBJECT pCentralDevice = NULL;
PDEVICE_OBJECT AIOWDMCreateCentralDevice(IN PDRIVER_OBJECT DriverObject);

NTSTATUS
AIOWDM4GetDeviceInformation(
    IN OUT PDEVICE_CONTEXT deviceContext
)
/*++
Routine Description:

    This function reads the PCI config space and make sure that it's our
    device and stores the device IDs in the device context. Should be done in the StartDevice.

Arguments:

    deviceContext     Pointer to our deviceContext

Return Value:

     None

--*/
{
    NTSTATUS            status = STATUS_SUCCESS;
    DECLSPEC_ALIGN(MEMORY_ALLOCATION_ALIGNMENT) UCHAR buffer[8];
    PPCI_COMMON_CONFIG  pPciConfig = (PPCI_COMMON_CONFIG)buffer;
    ULONG               bytesRead = 0;

    TraceEvents(TRACE_LEVEL_VERBOSE, DBG_INIT, "---> %!FUNC! Enter\n");

    PAGED_CODE();

    //
    // Get the BUS_INTERFACE_STANDARD for our device so that we can
    // read & write to PCI config space.
    //
    status = WdfFdoQueryForInterface(deviceContext->WdfDeviceObject,
        &GUID_BUS_INTERFACE_STANDARD,
        (PINTERFACE)&deviceContext->BusInterface,
        sizeof(BUS_INTERFACE_STANDARD),
        1, // Version
        NULL); //InterfaceSpecificData
    if (!NT_SUCCESS(status)) {
        TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "WdfFdoQueryForInterface failed %!STATUS!\n", status);
        deviceContext->bRealDevice = FALSE;
        return status;
    }

    RtlZeroMemory(buffer, sizeof(buffer));
    bytesRead = deviceContext->BusInterface.GetBusData(
        deviceContext->BusInterface.Context,
        PCI_WHICHSPACE_CONFIG, //READ
        buffer,
        FIELD_OFFSET(PCI_COMMON_CONFIG, VendorID),
        8);

    if (bytesRead != 8) {
        TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT,
            "GetBusData (8) failed =%d\n",
            bytesRead);
        return STATUS_INVALID_DEVICE_REQUEST;
    }

    deviceContext->DeviceID = pPciConfig->DeviceID;
    TraceEvents(TRACE_LEVEL_VERBOSE, DBG_INIT, "<--- %!FUNC! Exit\n");
    return status;
}

NTSTATUS
AIOWDM4CreateIRQQueue(
    _In_ WDFDEVICE Device
)
/*++

Routine Description:

     The I/O dispatch callbacks for the frameworks device object
     are configured in this function.

     A single default I/O Queue is configured for parallel request
     processing, and a driver context memory allocation is created
     to hold our structure QUEUE_CONTEXT.

Arguments:

    Device - Handle to a framework device object.

Return Value:

    VOID

--*/
{
    WDFQUEUE queue;
    NTSTATUS status;
    WDF_IO_QUEUE_CONFIG queueConfig;

    PAGED_CODE();

    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> %!FUNC! Enter");

    //
    // Configure a default queue so that requests that are not
    // configure-fowarded using WdfDeviceConfigureRequestDispatching to goto
    // other queues get dispatched here.
    //
    WDF_IO_QUEUE_CONFIG_INIT(
        &queueConfig,
        WdfIoQueueDispatchManual
    );


    status = WdfIoQueueCreate(
        Device,
        &queueConfig,
        WDF_NO_OBJECT_ATTRIBUTES,
        &queue
    );
    if (!NT_SUCCESS(status)) {
        TraceEvents(TRACE_LEVEL_ERROR, TRACE_QUEUE, "WdfIoQueueCreate failed %!STATUS!", status);
        return status;
    }

    PDEVICE_CONTEXT deviceContext = AIOWDM4DeviceGetContext(Device);
    deviceContext->IRQQueue = queue;

    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "<-- %!FUNC! Exit");

    return status;
}

NTSTATUS
DriverEntry(
    _In_ PDRIVER_OBJECT  DriverObject,
    _In_ PUNICODE_STRING RegistryPath
)
/*++

Routine Description:
    DriverEntry initializes the driver and is the first routine called by the
    system after the driver is loaded. DriverEntry specifies the other entry
    points in the function driver, such as EvtDevice and DriverUnload.

Parameters Description:

    DriverObject - represents the instance of the function driver that is loaded
    into memory. DriverEntry must initialize members of DriverObject before it
    returns to the caller. DriverObject is allocated by the system before the
    driver is loaded, and it is released by the system after the system unloads
    the function driver from memory.

    RegistryPath - represents the driver-specific path in the Registry.
    The function driver can use the path to store driver related data between
    reboots. The path does not store hardware-instance-specific data.

Return Value:

    STATUS_SUCCESS if successful,
    STATUS_UNSUCCESSFUL otherwise.

--*/
{
    WDF_DRIVER_CONFIG config;
    NTSTATUS Status;
    WDF_OBJECT_ATTRIBUTES attributes;


    //
    // Initialize WPP Tracing
    //
    WPP_INIT_TRACING(DriverObject, RegistryPath);

    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> %!FUNC! Entry");

    //
    // Register a cleanup callback so that we can call WPP_CLEANUP when
    // the framework driver object is deleted during driver unload.
    //
    WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
    attributes.EvtCleanupCallback = AIOWDM4EvtDriverContextCleanup;

    WDF_DRIVER_CONFIG_INIT(&config, AIOWDM4EvtDeviceAdd);

    Status = WdfDriverCreate(DriverObject,
        RegistryPath,
        &attributes,
        &config,
        WDF_NO_HANDLE
    );
    if (!NT_SUCCESS(Status)) {
        TraceEvents(TRACE_LEVEL_ERROR, TRACE_DRIVER, "WdfDriverCreate failed %!STATUS!", Status);
        WPP_CLEANUP(DriverObject);
        return Status;
    }

    // Create central device for global commands.
    UNICODE_STRING NtDeviceName;
    UNICODE_STRING Win32DeviceName;
    RtlInitUnicodeString(&NtDeviceName, DEVICE_DEVICE_NAME);
    RtlInitUnicodeString(&Win32DeviceName, AIOWDM_DEVICE_NAME);
    IoCreateSymbolicLink(&Win32DeviceName, &NtDeviceName);

    if (FALSE) // --J2H Debug
        if (pCentralDevice == NULL)
        {
            pCentralDevice = AIOWDMCreateCentralDevice(DriverObject);
            if (pCentralDevice == NULL)
                TraceEvents(TRACE_LEVEL_ERROR, TRACE_DRIVER, "AIOWDMCreateCentralDevice failed");
        }

    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "<-- %!FUNC! Exit");

    return Status;
}


NTSTATUS
AIOWDM4MapHWResources(
	_In_ OUT PDEVICE_CONTEXT deviceContext,
	_In_ WDFCMRESLIST  ResourcesRaw,
	_In_ WDFCMRESLIST  ResourcesTranslated
)
/*++
Routine Description:

	Gets the HW resources assigned by the bus driver and:
	1) Maps them to system address space.
	2) creates a WDFINTERRUPT object.

	Called during EvtDevicePrepareHardware callback.

	Three base address registers are supported by the PCIe-IDIO-24:
	1) 8311 Memory Mapped Base Address Register (BAR 0 at offset 10)
	2) 8311 I/O Mapped Base Address Register (BAR 1 at offset 14)
	3) DIO I/O Mapped Base Address Register (BAR 2 at offset 18)

Arguments:

	deviceContext     Pointer to our device extension
	ResourcesRaw - Pointer to list of raw resources passed to EvtDevicePrepareHardware callback
	ResourcesTranslated - Pointer to list of translated resources passed to EvtDevicePrepareHardware callback

Return Value:

	NTSTATUS

--*/
{
	PCM_PARTIAL_RESOURCE_DESCRIPTOR descriptor;
	ULONG       i;
	NTSTATUS    status = STATUS_SUCCESS;
	BOOLEAN     bResInterrupt = FALSE;
	ULONG       numberOfBARs = 0;

	UNREFERENCED_PARAMETER(ResourcesRaw);

	PAGED_CODE();

	TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> %!FUNC! Entry");

	for (i = 0; i < WdfCmResourceListGetCount(ResourcesTranslated); i++)
	{
		descriptor = WdfCmResourceListGetDescriptor(ResourcesTranslated, i);

		if ( ! descriptor )
		{
			TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "WdfResourceCmGetDescriptor");
			return STATUS_DEVICE_CONFIGURATION_ERROR;
		}

		switch (descriptor->Type)
		{

		case CmResourceTypePort:
			//
			// We will increment the BAR count only for valid resources. We will not count the private device types added by the PCI bus driver.
			//
			numberOfBARs++;

			TraceEvents
			(
				TRACE_LEVEL_VERBOSE, DBG_INIT,
				"I/O mapped CSR: (%x) Length: (%d)\n",
				descriptor->u.Port.Start.LowPart,
				descriptor->u.Port.Length
			);

			//
			// The resources are listed in the same order as the BARs in config space
			//
			// The port is in I/O space on this device.
			// We should use READ_PORT_Xxx, and WRITE_PORT_Xxx routines to read or write to the port.
			//
			// switch ( numberOfBARs )
			// {
			// default:
				TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "unexpected I/O mapped BAR found\n");
				status = STATUS_DEVICE_CONFIGURATION_ERROR;
				return status;
			// break;
			// }
		break;

		case CmResourceTypeMemory:
			numberOfBARs++;

			switch ( numberOfBARs )
			{
			case 1:
				TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "Found memory mapped \"PLX\" status BAR@ %llx\n", descriptor->u.Port.Start.QuadPart);
				deviceContext->StatusBar = BarFromResource(*descriptor);
				/*
				deviceContext->StatusBar.BarType = btMemory;
				deviceContext->StatusBar.Base = `Map`(PUCHAR(UNATIVE(descriptor->u.Port.Start.QuadPart)));
				deviceContext->StatusBar.Size = descriptor->u.Port.Length;
				*/
			break;
			case 2:
				TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "Found memory mapped aux BAR:(%llx)\n", descriptor->u.Memory.Start.QuadPart);
				deviceContext->AuxBar = BarFromResource(*descriptor);
			break;
			case 3:
				TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INIT, "Found memory mapped main BAR @ %llx\n", descriptor->u.Port.Start.QuadPart);
				deviceContext->MainBar = BarFromResource(*descriptor);
				/*
				deviceContext->MainBar.BarType = btMemory;
				deviceContext->MainBar.Base = `Map`(PUCHAR(UNATIVE(descriptor->u.Port.Start.QuadPart)));
				deviceContext->MainBar.Size = descriptor->u.Port.Length;
				*/
			break;
			default:
				TraceEvents
				(
					TRACE_LEVEL_INFORMATION, DBG_INIT,
					"Unexpected extra Memory BAR: (%llx) Length: (%d)\n",
					descriptor->u.Memory.Start.QuadPart,
					descriptor->u.Memory.Length
				);
				//status = STATUS_DEVICE_CONFIGURATION_ERROR;
				//return status;
			break;
			}
		break;

		case CmResourceTypeInterrupt:
		{
			WDF_INTERRUPT_CONFIG interruptConfig;
			TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "CmResourceTypeInterrupt found");
			//
			// Create WDFINTERRUPT object.
			//
			WDF_INTERRUPT_CONFIG_INIT
			(
				&interruptConfig,
				AIOWDM4EvtInterruptIsr,
				AIOWDM4EvtInterruptDpc
			);

			//
			// These first two callbacks will be called at DIRQL.  Their job is to
			// enable and disable interrupts.
			//
			interruptConfig.EvtInterruptEnable = AIOWDM4EvtInterruptEnable;
			interruptConfig.EvtInterruptDisable = AIOWDM4EvtInterruptDisable;
			interruptConfig.InterruptTranslated = descriptor;
			interruptConfig.InterruptRaw = WdfCmResourceListGetDescriptor(ResourcesRaw, i);

			status = WdfInterruptCreate
			(
				deviceContext->WdfDeviceObject, // --J2H change confirm
				&interruptConfig,
				WDF_NO_OBJECT_ATTRIBUTES,
				&deviceContext->InterruptObject
			);

			if ( ! NT_SUCCESS(status) )
			{
				TraceEvents
				(
					TRACE_LEVEL_ERROR, DBG_INIT,
					"WdfInterruptCreate failed: %!STATUS!\n", status
				);
				return status;
			}

			AIOWDM4CreateIRQQueue(deviceContext->WdfDeviceObject);

			bResInterrupt = TRUE;

			TraceEvents
			(
				TRACE_LEVEL_INFORMATION, TRACE_DRIVER,
				"Interrupt level: 0x%0x, Vector: 0x%0x\n",
				descriptor->u.Interrupt.Level,
				descriptor->u.Interrupt.Vector
			);
		}
		break;

		default:
			//
			// This could be device-private type added by the PCI bus driver. We
			// shouldn't filter this or change the information contained in it.
			//
			TraceEvents(TRACE_LEVEL_VERBOSE, DBG_INIT, "Unhandled resource type (0x%x)\n", descriptor->Type);
		break;
		}

	}

	//
	// Read additional info from PCIe-IDIO-24 such as DeviceID
	//
	status = AIOWDM4GetDeviceInformation(deviceContext);
	if ( status != STATUS_SUCCESS )
	{
		TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "Central Device found?");
	}

	TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "<-- %!FUNC! Exit");

	return status;
}

NTSTATUS
AIOWDM4UnmapHWResources(
	_In_ OUT PDEVICE_CONTEXT deviceContext
)
/*++
Routine Description:

	Disconnect the interrupt and unmap all the memory and I/O resources.

Arguments:

	deviceContext     Pointer to our deviceContext

Return Value:

	 None

--*/
{
	PAGED_CODE();



	DelBar(deviceContext->StatusBar);
	DelBar(deviceContext->MainBar);
	DelBar(deviceContext->AuxBar);

	return STATUS_SUCCESS;



	//UNREFERENCED_PARAMETER(deviceContext);

	//
	// Free hardware resources
	//
	// -- J2H TODO
	/*
	if (deviceContext->CSRAddress)
	{
		MmUnmapIoSpace(deviceContext->CSRAddress, NIC_MAP_IOSPACE_LENGTH);
		deviceContext->CSRAddress = NULL;
	}

	if (deviceContext->MappedPorts)
	{
		MmUnmapIoSpace(deviceContext->IoBaseAddress, deviceContext->IoRange);
		deviceContext->IoBaseAddress = NULL;
	}
	*/

	//return STATUS_SUCCESS;
}



EVT_WDF_DEVICE_PREPARE_HARDWARE AIOWDM4EvtPrepareHardware;

NTSTATUS
AIOWDM4EvtPrepareHardware(
    WDFDEVICE Device,
    WDFCMRESLIST ResourcesRaw,
    WDFCMRESLIST ResourcesTranslated
)
{
    UNREFERENCED_PARAMETER(ResourcesRaw);
    UNREFERENCED_PARAMETER(ResourcesTranslated);

    NTSTATUS     status = STATUS_SUCCESS;
    PDEVICE_CONTEXT    deviceContext = NULL;

    UNREFERENCED_PARAMETER(ResourcesTranslated);

    PAGED_CODE();

    TraceEvents(TRACE_LEVEL_INFORMATION, DBG_PNP, "--> %!FUNC! Entry\n");

    deviceContext = AIOWDM4DeviceGetContext(Device);

    status = AIOWDM4MapHWResources(deviceContext, ResourcesRaw, ResourcesTranslated);
    if (!NT_SUCCESS(status)) {
        TraceEvents(TRACE_LEVEL_ERROR, DBG_PNP, "AIOWDM4MapHWResources failed: %!STATUS!\n", status);
    }
    else
    {
        //
        // Create a device interface so that applications can find and talk
        // to us.
        //
        status = WdfDeviceCreateDeviceInterface(Device, &GUID_DEVINTERFACE_AIOWDM4, NULL); // ReferenceString
        if (!NT_SUCCESS(status)) {
            TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "WdfDeviceCreateDeviceInterface failed %!STATUS!\n", status);
            return status;
        }
    }

    TraceEvents(TRACE_LEVEL_INFORMATION, DBG_PNP, "<-- %!FUNC! Exit\n");
    return status;
}

EVT_WDF_DEVICE_RELEASE_HARDWARE AIOWDM4EvtReleaseHardware;

NTSTATUS
AIOWDM4EvtReleaseHardware(
    WDFDEVICE Device,
    WDFCMRESLIST ResourcesTranslated
)
{
    UNREFERENCED_PARAMETER(Device);
    UNREFERENCED_PARAMETER(ResourcesTranslated);
    return STATUS_SUCCESS;
    /// TBD
}



NTSTATUS
AIOWDM4EvtDeviceAdd(
    _In_    WDFDRIVER       Driver,
    _Inout_ PWDFDEVICE_INIT DeviceInit
)
/*++
Routine Description:

    EvtDeviceAdd is called by the framework in response to AddDevice
    call from the PnP manager. We create and initialize a device object to
    represent a new instance of the device.

Arguments:

    Driver - Handle to a framework driver object created in DriverEntry

    DeviceInit - Pointer to a framework-allocated WDFDEVICE_INIT structure.

Return Value:

    NTSTATUS

--*/
{
    NTSTATUS status;

    UNREFERENCED_PARAMETER(Driver);

    PAGED_CODE();

    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> %!FUNC! Enter");

    /* initialize the DeviceInit structure "WDFDEVICE_INIT": "DeviceInit" passed into this function
    */

    WDF_PNPPOWER_EVENT_CALLBACKS  pnpPowerCallbacks;
    WDF_PNPPOWER_EVENT_CALLBACKS_INIT(&pnpPowerCallbacks);
    pnpPowerCallbacks.EvtDevicePrepareHardware = AIOWDM4EvtPrepareHardware; // initialize on power-up
    pnpPowerCallbacks.EvtDeviceReleaseHardware = AIOWDM4EvtReleaseHardware; // de-init anything from above
    WdfDeviceInitSetPnpPowerEventCallbacks(DeviceInit, &pnpPowerCallbacks);

    WDF_OBJECT_ATTRIBUTES myDeviceObjectAttributes;

    WDF_OBJECT_ATTRIBUTES_INIT(&myDeviceObjectAttributes);
    WDF_OBJECT_ATTRIBUTES_SET_CONTEXT_TYPE(&myDeviceObjectAttributes, DEVICE_CONTEXT);

    // TODO: Am I supposed to allocate something now??

    status = AIOWDM4CreateDevice(DeviceInit);

    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "<-- %!FUNC! Exit, Status: %!STATUS!", status);

    return status;
}

VOID
AIOWDM4EvtDriverContextCleanup(
    _In_ WDFOBJECT DriverObject
)
/*++
Routine Description:

    Free all the resources allocated in DriverEntry.

Arguments:

    DriverObject - handle to a WDF Driver object.

Return Value:

    VOID.

--*/
{
    UNREFERENCED_PARAMETER(DriverObject);

    PAGED_CODE();
    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> %!FUNC! excecuting");

    UNICODE_STRING Win32DeviceName;

    RtlInitUnicodeString(&Win32DeviceName, AIOWDM_DEVICE_NAME);
    IoDeleteSymbolicLink(&Win32DeviceName);

    //
    // Stop WPP Tracing
    //
    WPP_CLEANUP(WdfDriverWdmGetDriverObject((WDFDRIVER)DriverObject));

}
