/*++

Copyright (c) Microsoft Corporation.  All rights reserved.

    THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
    KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
    PURPOSE.

Module Name:

    ISRDPC.C

Abstract:

    Contains routine to handle interrupts, interrupt DPCs and WatchDogTimer DPC

Environment:

    Kernel mode

--*/

#include "driver.h"
#include "IsrDpc.tmh"


#ifdef ALLOC_PRAGMA

#endif


#define AIOWDM4_INTERRUPT_DISABLED(_adapter) \
   ((READ_PORT_UCHAR(_adapter->StatusBar.Base + IRQEnableOffset) & maskEnableIRQ) != 0x00)

#define AIOWDM4_INTERRUPT_ACTIVE(_adapter) \
      (READ_PORT_UCHAR(_adapter->MainBar.Base + IRQStatusOffset) != 0x00) 


#define AIOWDM4_ACK_INTERRUPT(_adapter, _value) { \
   *_value = READ_PORT_ULONG(_adapter->MainBar.Base + COSIRQLatchOffset); \
   WRITE_PORT_ULONG(_adapter->MainBar.Base + COSIRQLatchOffset, *_value); }

BOOLEAN
AIOWDM4EvtInterruptIsr(
    IN WDFINTERRUPT Interrupt,
    IN ULONG        MessageID
)
/*++
Routine Description:
    Interrupt handler for the device.
Arguments:
    Interupt - Address of the framework interrupt object
    MessageID -
Return Value:
     TRUE if our device is interrupting, FALSE otherwise.`
--*/
{
    BOOLEAN    InterruptRecognized = FALSE;
    PDEVICE_CONTEXT  deviceContext = NULL;

    UNREFERENCED_PARAMETER(MessageID);

    deviceContext = AIOWDM4DeviceGetContext(WdfInterruptGetDevice(Interrupt));

    //
    // We process the interrupt if it's not disabled and it's active
    //
    ULONG IntStatus = READ_PORT_UCHAR((PUCHAR)(deviceContext->MainBar.Base + IRQStatusOffset));
    TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INTERRUPT, "I AM (%04x)=>%02lx", (ULONG)(UNATIVE)(deviceContext->MainBar.Base + IRQStatusOffset), IntStatus);
 
    UCHAR En = READ_PORT_UCHAR(deviceContext->StatusBar.Base + IRQEnableOffset);
    TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INTERRUPT, "8311 IS (%04x)=>%02hx", (ULONG)(UNATIVE)(deviceContext->StatusBar.Base + IRQEnableOffset), En);

    if (IntStatus && En) // if an IRQ is fired and it was me
    {
        TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INTERRUPT, "--> %!FUNC! enter IAm\n");
        InterruptRecognized = TRUE;

        //
        // Disable the interrupt (will be re-enabled in AIOWDM4EvtInterruptDpc)
        //
        WRITE_PORT_UCHAR(deviceContext->StatusBar.Base + IRQEnableOffset, maskDisableIRQ);

        //
        // Acknowledge the interrupt(s) and get the interrupt status
        //
        // AIOWDM4_ACK_INTERRUPT(deviceContext, &IntStatus);
        IntStatus = READ_PORT_ULONG((PULONG)(deviceContext->MainBar.Base + COSIRQLatchOffset));
        WRITE_PORT_ULONG((PULONG)(deviceContext->MainBar.Base + COSIRQLatchOffset), IntStatus);

        WdfInterruptQueueDpcForIsr(Interrupt);
    }

    TraceEvents(TRACE_LEVEL_INFORMATION, DBG_INTERRUPT, "<-- %!FUNC! exit\n");

    return InterruptRecognized;
}


VOID InCOSData(PDEVICE_CONTEXT pdx, PUCHAR pDst, size_t COSSize);

VOID
AIOWDM4HandleInterrupt(PDEVICE_CONTEXT deviceContext)
/*++
Routine Description:
    Notify IRQQueue an IRQ occurred
Arguments:
    deviceContext - Associated device context.
Return Value:
--*/
{
    UNREFERENCED_PARAMETER(deviceContext);
    TraceEvents(TRACE_LEVEL_INFORMATION, DBG_DPC, "--> %!FUNC! Enter");

    NTSTATUS status;
    ULONG_PTR infolen=0;
    WDFREQUEST notifyRequest;
    PULONG  bufferPointer;
//    ULONG PendingIRQWaits;

    //WDF_IO_QUEUE_STATE queueStatus;
    //queueStatus = WdfIoQueueGetState(
    //    deviceContext->IRQQueue,
    //    NULL,
    //    &PendingIRQWaits
    //);

    if (deviceContext->HasPendingIrp == FALSE)
    {
        deviceContext->InterruptOccurred = TRUE;
        // And now... NOTIFY the user about the event.  We do this just
        // by completing the dequeued Request.

        TraceEvents(TRACE_LEVEL_INFORMATION, DBG_DPC, "<-- %!FUNC! NO PENDING WAITS Exit");
    }
    else
    //while (deviceContext->InterruptRequested--)
    {
        status = WdfIoQueueRetrieveNextRequest(deviceContext->IRQQueue, &notifyRequest);
        if (!NT_SUCCESS(status)) {
            // Nope!  We were NOT able to successfully remove a Request from the
            // notification queue.  Well, perhaps there aren't any right now.
            // Whatever... not much we can do in this example about this.
            TraceEvents(TRACE_LEVEL_INFORMATION, DBG_DPC, "<-- %!FUNC! PENDING WAITS FAILURE %d Exit\n", status);
            return;
        }

        //
        // We've successfully removed a Request from the queue of pending 
        // notification IOCTLs.
        // 
        //
        // Get a a pointer to the output buffer that was passed-in with the user
        // notification IOCTL.  We'll use this to return additional infolen about
        // the event. The minimum size Output Buffer that we need is sizeof(LONG).
        // We don't need this call to return us what the actual size is.
        // 
        status = WdfRequestRetrieveOutputBuffer(notifyRequest,
            sizeof(LONG),                                           //--J2H TODO: Change to COSSize
            (PVOID*)&bufferPointer,
            NULL);
        if (!NT_SUCCESS(status))
        if (TRUE)
        {
            // Complete the IOCTL_OSR_INVERT_NOTIFICATION with success, but
            // indicate that we're not returning any additional information.
            status = STATUS_SUCCESS;
            infolen = 0;
            WdfRequestComplete(notifyRequest, status);
            deviceContext->HasPendingIrp = FALSE;
            TraceEvents(TRACE_LEVEL_INFORMATION, DBG_DPC, "nonCoS Request Completed");
        }
        else { // CoSWaitForIRQ
            // We successfully retrieved a Request from the notification Queue
            // AND we retrieved an output buffer into which to return some
            // additional information.
            InCOSData(deviceContext, (PUCHAR)bufferPointer, deviceContext->COSSize);

            //
            // Complete the IOCTL_IRQ_WAIT_BEGIN with success, and indicate
            // we're returning CoS data in the user's OutBuffer
            // 
            status = STATUS_SUCCESS;
            infolen = deviceContext->COSSize;
            WdfRequestCompleteWithInformation(notifyRequest, status, infolen);
            deviceContext->HasPendingIrp = FALSE;
            TraceEvents(TRACE_LEVEL_INFORMATION, DBG_DPC, "CoS Request Completed");
        }

        // And now... NOTIFY the user about the event.  We do this just
        // by completing the dequeued Request.       
    }
    TraceEvents(TRACE_LEVEL_INFORMATION, DBG_DPC, "<-- %!FUNC! Exit");
}

VOID
AIOWDM4EvtInterruptDpc(
    IN WDFINTERRUPT WdfInterrupt,
    IN WDFOBJECT    WdfDevice
)
/*++
Routine Description:
    DPC callback for ISR.
Arguments:
    WdfInterrupt - Handle to the framework interrupt object
    WdfDevice - Associated device object.
Return Value:
--*/
{
    UNREFERENCED_PARAMETER(WdfInterrupt);
    TraceEvents(TRACE_LEVEL_INFORMATION, DBG_DPC, "--> %!FUNC!\n");

    PDEVICE_CONTEXT deviceContext = AIOWDM4DeviceGetContext(WdfDevice);

    //WdfSpinLockAcquire(deviceContext->RcvLock);

    AIOWDM4HandleInterrupt(deviceContext);

    //WdfSpinLockRelease(deviceContext->RcvLock);

    //
    // Re-enable the interrupt (disabled in MPIsr)
    //
    WRITE_PORT_UCHAR(deviceContext->StatusBar.Base + IRQEnableOffset, maskEnableIRQ);

    TraceEvents(TRACE_LEVEL_INFORMATION, DBG_DPC, "<-- %!FUNC!\n");
}


NTSTATUS
AIOWDM4EvtInterruptEnable(
    IN WDFINTERRUPT  Interrupt,
    IN WDFDEVICE     AssociatedDevice
)
/*++

Routine Description:

    This event is called when the Framework moves the device to D0, and after
    EvtDeviceD0Entry.  The driver should enable its interrupt here.

    This function will be called at the device's assigned interrupt
    IRQL (DIRQL.)

Arguments:

    Interrupt - Handle to a Framework interrupt object.

    AssociatedDevice - Handle to a Framework device object.

Return Value:

    BOOLEAN - TRUE indicates that the interrupt was successfully enabled.

--*/
{
    PDEVICE_CONTEXT           deviceContext;
    UNREFERENCED_PARAMETER(Interrupt);
    TraceEvents(TRACE_LEVEL_VERBOSE, DBG_PNP, "--> %!FUNC!\n");

    deviceContext = AIOWDM4DeviceGetContext(AssociatedDevice);

    WRITE_PORT_UCHAR(deviceContext->StatusBar.Base + IRQEnableOffset, maskEnableIRQ);
    //WdfInterruptSynchronize(Interrupt, AIOWDM4EnableInterrupt, deviceContext);

    TraceEvents(TRACE_LEVEL_VERBOSE, DBG_PNP, "<-- %!FUNC!\n");

    return STATUS_SUCCESS;
}

NTSTATUS
AIOWDM4EvtInterruptDisable(
    IN WDFINTERRUPT  Interrupt,
    IN WDFDEVICE     AssociatedDevice
)
/*++

Routine Description:

    This event is called before the Framework moves the device to D1, D2 or D3
    and before EvtDeviceD0Exit.  The driver should disable its interrupt here.

    This function will be called at the device's assigned interrupt
    IRQL (DIRQL.)

Arguments:

    Interrupt - Handle to a Framework interrupt object.

    AssociatedDevice - Handle to a Framework device object.

Return Value:

    STATUS_SUCCESS -  indicates success.

--*/
{
    UNREFERENCED_PARAMETER(Interrupt);
    PDEVICE_CONTEXT deviceContext;

    TraceEvents(TRACE_LEVEL_VERBOSE, DBG_PNP, "--> %!FUNC!\n");

    deviceContext = AIOWDM4DeviceGetContext(AssociatedDevice);
    WRITE_PORT_UCHAR(deviceContext->StatusBar.Base + IRQEnableOffset, maskDisableIRQ);

    TraceEvents(TRACE_LEVEL_VERBOSE, DBG_PNP, "<-- %!FUNC!\n");

    return STATUS_SUCCESS;
}
